# Copyright (c) Alibaba, Inc. and its affiliates.
# Some codes are borrowed from AllenNLP.
# https://github.com/allenai/allennlp/blob/HEAD/allennlp/common/logging.py
# Copyright (c) AI2 AllenNLP. Licensed under the Apache License, Version 2.0.

import logging
import os
import sys
from logging import Filter
from os import PathLike
from typing import Union

from modelscope.utils.logger import get_logger


class AdaSeqLogger(logging.Logger):
    """
    A custom subclass of 'logging.Logger' that keeps a set of messages to
    implement {debug,info,etc.}_once() methods.
    """

    def __init__(self, name):
        super().__init__(name)
        self._seen_msgs = set()

    def debug_once(self, msg, *args, **kwargs):  # noqa: D102
        if msg not in self._seen_msgs:
            self.debug(msg, *args, **kwargs)
            self._seen_msgs.add(msg)

    def info_once(self, msg, *args, **kwargs):  # noqa: D102
        if msg not in self._seen_msgs:
            self.info(msg, *args, **kwargs)
            self._seen_msgs.add(msg)

    def warning_once(self, msg, *args, **kwargs):  # noqa: D102
        if msg not in self._seen_msgs:
            self.warning(msg, *args, **kwargs)
            self._seen_msgs.add(msg)

    def error_once(self, msg, *args, **kwargs):  # noqa: D102
        if msg not in self._seen_msgs:
            self.error(msg, *args, **kwargs)
            self._seen_msgs.add(msg)

    def critical_once(self, msg, *args, **kwargs):  # noqa: D102
        if msg not in self._seen_msgs:
            self.critical(msg, *args, **kwargs)
            self._seen_msgs.add(msg)


logging.setLoggerClass(AdaSeqLogger)
logger = logging.getLogger(__name__)


class ErrorFilter(Filter):
    """
    Filters out everything that is at the ERROR level or higher. This is meant to be used
    with a stdout handler when a stderr handler is also configured. That way ERROR
    messages aren't duplicated.
    """

    def filter(self, record):  # noqa: D102
        return record.levelno < logging.ERROR


def prepare_global_logging(log_level: int = logging.INFO) -> None:
    """
    Prepare global logging.
    """

    # clear modelscop logger handlers to remove dumlicate logs.
    get_logger().handlers.clear()

    root_logger = logging.getLogger()

    # create handlers
    formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(name)s - %(message)s')
    stdout_handler: logging.Handler = logging.StreamHandler(sys.stdout)
    stderr_handler: logging.Handler = logging.StreamHandler(sys.stderr)

    for handler in [stdout_handler, stderr_handler]:
        handler.setFormatter(formatter)

    # Remove the already set handlers in root logger.
    # Not doing this will result in duplicate log messages
    root_logger.handlers.clear()

    if os.environ.get('ADASEQ_DEBUG'):
        LEVEL = logging.DEBUG
    else:
        LEVEL = log_level

    stdout_handler.setLevel(LEVEL)
    stdout_handler.addFilter(ErrorFilter())  # Make sure errors only go to stderr
    stderr_handler.setLevel(logging.ERROR)
    root_logger.setLevel(LEVEL)

    # put all the handlers on the root logger
    root_logger.addHandler(stdout_handler)
    root_logger.addHandler(stderr_handler)


def prepare_logging(
    work_dir: Union[str, PathLike],
    rank: int = 0,
    world_size: int = 1,
    log_level: int = logging.INFO,
) -> None:
    """
    Prepare logging for training, log to file in `work_dir`.
    """
    root_logger = logging.getLogger()

    # create handlers
    if world_size == 1:
        log_file = os.path.join(work_dir, 'out.log')
        formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(name)s - %(message)s')
    else:
        log_file = os.path.join(work_dir, f'out_worker{rank}.log')
        formatter = logging.Formatter(
            f'{rank} | %(asctime)s - %(levelname)s - %(name)s - %(message)s'
        )
    file_handler: logging.Handler = logging.FileHandler(log_file)
    stdout_handler: logging.Handler = logging.StreamHandler(sys.stdout)
    stderr_handler: logging.Handler = logging.StreamHandler(sys.stderr)

    handler: logging.Handler
    for handler in [file_handler, stdout_handler, stderr_handler]:
        handler.setFormatter(formatter)

    # Remove the already set handlers in root logger.
    # Not doing this will result in duplicate log messages
    root_logger.handlers.clear()

    if os.environ.get('ADASEQ_DEBUG'):
        LEVEL = logging.DEBUG
    else:
        LEVEL = log_level

    file_handler.setLevel(LEVEL)
    stdout_handler.setLevel(LEVEL)
    stdout_handler.addFilter(ErrorFilter())  # Make sure errors only go to stderr
    stderr_handler.setLevel(logging.ERROR)
    root_logger.setLevel(LEVEL)

    # put all the handlers on the root logger
    root_logger.addHandler(file_handler)
    if rank == 0:
        root_logger.addHandler(stdout_handler)
        root_logger.addHandler(stderr_handler)



# Copyright (c) Alibaba, Inc. and its affiliates.
from typing import Any, Dict, List, Union

import numpy as np
from modelscope.preprocessors.builder import PREPROCESSORS
from modelscope.utils.constant import Fields

from adaseq.metainfo import Preprocessors

from .nlp_preprocessor import NLPPreprocessor


@PREPROCESSORS.register_module(Fields.nlp, module_name=Preprocessors.span_extraction_preprocessor)
class SpanExtracionPreprocessor(NLPPreprocessor):
    """Preprocessor of span-based model.
    span targets are processed into `span_labels`
    """

    def __init__(self, model_dir: str, **kwargs):
        super().__init__(model_dir, return_offsets=True, **kwargs)

    def __call__(self, data: Union[str, List, Dict]) -> Dict[str, Any]:
        """prepare inputs for span-based model."""

        output = super().__call__(data)

        if not (isinstance(data, dict) and 'spans' in data):
            return output

        # origin sequence length
        length = len(output['tokens']['mask']) - 2 * int(self.add_special_tokens)
        # calculate span labels
        span_labels = np.zeros([length, length])
        for span in data['spans']:
            # self.label_to_id doesn't have non-entity label,
            # we set index 0 as non-entity label, so we add 1 to type_id
            type_id = self.label_to_id[span['type']] + 1
            span_labels[span['start']][span['end'] - 1] = type_id
        output['span_labels'] = span_labels
        return output



# Copyright (c) Alibaba, Inc. and its affiliates.
import argparse
import os
from typing import Optional

from modelscope.utils.config import Config

from adaseq.commands.subcommand import Subcommand
from adaseq.data.data_collators.base import build_data_collator
from adaseq.data.dataset_manager import DatasetManager
from adaseq.data.preprocessors.nlp_preprocessor import build_preprocessor
from adaseq.metainfo import Trainers
from adaseq.training import build_trainer
from adaseq.utils.checks import ConfigurationError
from adaseq.utils.file_utils import is_empty_dir


class Test(Subcommand):
    """
    usage: adaseq test [-h] -w WORK_DIR [-d DEVICE] [-ckpt CHECKPOINT_PATH]

    optional arguments:
      -h, --help            show this help message and exit
      -w WORK_DIR, --work_dir WORK_DIR
                            directory to load config and checkpoint
      -d DEVICE, --device DEVICE
                            device name
      -ckpt CHECKPOINT_PATH, --checkpoint_path CHECKPOINT_PATH
                            model checkpoint
    """

    @classmethod
    def add_subparser(cls, parser: argparse._SubParsersAction) -> argparse.ArgumentParser:
        """Add testing arguments parser"""
        subparser = parser.add_parser('test', help='test with a model checkpoint')
        subparser.add_argument(
            '-w', '--work_dir', required=True, help='directory to load config and checkpoint'
        )
        subparser.add_argument('-d', '--device', default='gpu', help='device name')
        subparser.add_argument('-ckpt', '--checkpoint_path', default=None, help='model checkpoint')

        subparser.set_defaults(func=test_model_from_args)
        return subparser


def test_model_from_args(args: argparse.Namespace):  # noqa: D103
    test_model(
        work_dir=args.work_dir,
        device=args.device,
        checkpoint_path=args.checkpoint_path,
    )


def test_model(
    work_dir: str,
    device: str = 'gpu',
    checkpoint_path: Optional[str] = None,
) -> None:
    """
    Train a model from config file.
    You can mannualy call this function in a python script for debugging.
    """
    config = Config.from_file(os.path.join(work_dir, 'config.yaml'))
    checkpoint_path = checkpoint_path or os.path.join(work_dir, 'best_model.pth')

    if not os.path.exists(work_dir) and not is_empty_dir(work_dir):
        raise ConfigurationError(f'`work_dir` ({work_dir}) do not exists or is not empty.')

    # build datasets via `DatasetManager`
    dm = DatasetManager.from_config(task=config.task, **config.dataset)
    # build preprocessor with config and labels
    preprocessor = build_preprocessor(config.preprocessor, labels=dm.labels)

    # Finally, get `id_to_label` for model.
    assert config.model.id_to_label == preprocessor.id_to_label

    # build `DataCollator` from config and tokenizer.
    collator_config = config.data_collator
    if isinstance(collator_config, str):
        collator_config = dict(type=collator_config)
    data_collator = build_data_collator(preprocessor.tokenizer, collator_config)

    trainer = build_trainer(
        config.safe_get('train.trainer', Trainers.default_trainer),
        cfg_file=config.filename,
        seed=config.safe_get('experiment.seed', 42),
        device=device,
        work_dir=work_dir,
        dataset_manager=dm,
        data_collator=data_collator,
        preprocessor=preprocessor,
    )
    trainer.test(checkpoint_path)



# Copyright (c) Alibaba, Inc. and its affiliates.

import json
import logging
import re
from typing import Any, Dict, List, Optional, Tuple, Union

from modelscope.trainers.optimizer.builder import OPTIMIZERS
from modelscope.utils.config import ConfigDict
from modelscope.utils.registry import build_from_cfg, default_group
from torch import nn

logger = logging.getLogger(__name__)


def build_optimizer(model: nn.Module, cfg: ConfigDict, default_args: dict = None):
    """Build optimizer from config"""
    cfg = cfg.copy()  # proctect the base config is not modified

    if hasattr(model, 'module'):
        model = model.module  # type ignore

    # build parameter groups with different kwargs
    groups = cfg.pop('param_groups', None)
    param_groups = make_parameter_groups(model.named_parameters(), groups)  # type: ignore

    if default_args is None:
        default_args = {}
    default_args['params'] = param_groups

    return build_from_cfg(cfg, OPTIMIZERS, group_key=default_group, default_args=default_args)


# The implementation is adopted from AllenNLP and modified.
# https://github.com/allenai/allennlp/blob/main/allennlp/training/optimizers.py
# Licensed under the Apache License, Version 2.0.
def make_parameter_groups(
    model_parameters: List[Tuple[str, nn.Parameter]],  # type: ignore
    groups: Optional[List[Dict[str, Any]]] = None,
) -> Union[List[Dict[str, Any]], List[nn.Parameter]]:  # type: ignore
    """
    Takes a list of model parameters with associated names (typically coming from something like
    `model.named_parameters()`), along with a grouping (as specified below), and prepares them to be passed
    to the `__init__` function of a `torch.Optimizer`.  This means separating the parameters into
    groups with the given regexes, and prepping whatever keyword arguments are given for those
    regexes in `groups`.
    `groups` contains something like:
    ```
    [
        { 'regex': "transformer_model", 'lr': 1e-3 },
        { 'regex': ['re1', 're2', 're3'], 'lr': 1e-4 }
    ]
    ```
    All of key-value pairs specified in each of these dictionaries will passed passed as-is
    to the optimizer, with the exception of a dictionaries that specify `requires_grad` to be `False`:
    ```
    [
        ...
        {'regex': 'regex', 'requires_grad': False }
    ]
    ```
    When a parameter group has `{"requires_grad": False}`, the gradient on all matching parameters
    will be disabled and that group will be dropped so that it's not actually passed to the optimizer.
    Ultimately, the return value of this function is in the right format to be passed directly
    as the `params` argument to a pytorch `Optimizer`.
    If there are multiple groups specified, this is a list of dictionaries, where each
    dict contains a "parameter group" and groups specific options, e.g., {'params': [list of
    parameters], 'lr': 1e-3, ...}.  Any config option not specified in the additional options (e.g.
    for the default group) is inherited from the top level arguments given in the constructor.  See:
    <https://pytorch.org/docs/0.3.0/optim.html?#per-parameter-options>.  See also our
    `test_optimizer_parameter_groups` test for an example of how this works in this code.
    The dictionary's return type is labeled as `Any`, because it can be a `List[torch.nn.Parameter]`
    (for the "params" key), or anything else (typically a float) for the other keys.
    """
    if groups:
        # reformat eache group to ('regex', {})
        allennlp_groups = list()
        for k, group_regexes in enumerate(groups):
            regexes = group_regexes.pop('regex')
            if isinstance(regexes, str):
                regexes = [regexes]
            if not isinstance(regexes, list):
                raise ValueError(f'Unsopported regex: {regexes}')
            allennlp_groups.append((regexes, group_regexes))
        groups = allennlp_groups

        # In addition to any parameters that match group specific regex,
        # we also need a group for the remaining "default" group.
        # Those will be included in the last entry of parameter_groups.
        parameter_groups: Union[List[Dict[str, Any]], List[nn.Parameter]] = [  # type: ignore
            {'params': []} for _ in range(len(groups) + 1)
        ]
        # add the group specific kwargs
        for k in range(len(groups)):
            parameter_groups[k].update(groups[k][1])  # type: ignore

        regex_use_counts: Dict[str, int] = {}
        parameter_group_names: List[set] = [set() for _ in range(len(groups) + 1)]
        for name, param in model_parameters:
            # Determine the group for this parameter.
            group_index = None
            for k, group_regexes in enumerate(groups):
                for regex in group_regexes[0]:  # type: ignore
                    if regex not in regex_use_counts:
                        regex_use_counts[regex] = 0
                    if re.search(regex, name):
                        if group_index is not None and group_index != k:
                            raise ValueError(
                                '{} was specified in two separate parameter groups'.format(name)
                            )
                        group_index = k
                        regex_use_counts[regex] += 1

            if group_index is not None:
                parameter_groups[group_index]['params'].append(param)
                parameter_group_names[group_index].add(name)
            else:
                # the default group
                parameter_groups[-1]['params'].append(param)
                parameter_group_names[-1].add(name)

        # find and remove any groups with 'requires_grad = False'
        no_grad_group_indices: List[int] = []
        for k, (names, group) in enumerate(zip(parameter_group_names, parameter_groups)):
            if group.get('requires_grad') is False:
                no_grad_group_indices.append(k)
                logger.info(
                    'Disabling gradient for the following parameters: %s',
                    json.dumps(sorted(names), indent=2),
                )
                for param in group['params']:
                    param.requires_grad_(False)

                # warn about any other unused options in that group.
                unused_options = {
                    key: val for key, val in group.items() if key not in ('params', 'requires_grad')
                }
                if unused_options:
                    logger.warning(
                        'Ignoring unused options %s for %s',
                        unused_options,
                        json.dumps(sorted(names), indent=2),
                    )
        parameter_group_names = [
            names
            for (k, names) in enumerate(parameter_group_names)
            if k not in no_grad_group_indices
        ]
        parameter_groups = [
            group for (k, group) in enumerate(parameter_groups) if k not in no_grad_group_indices
        ]

        # log the remaining parameter groups
        logger.info('Done constructing parameter groups.')
        for k in range(len(parameter_groups)):
            group_options = {
                key: val for key, val in parameter_groups[k].items() if key != 'params'
            }
            name_string = json.dumps(sorted(parameter_group_names[k]), indent=2)
            logger.info('Group %s: %s, %s', k, group_options, name_string)

        # check for unused regex
        for regex, count in regex_use_counts.items():
            if count == 0:
                logger.warning(
                    'When constructing parameter groups, %s does not match any parameter name',
                    regex,
                )
    else:
        parameter_groups = [param for name, param in model_parameters]

    # Log the number of parameters to optimize
    num_parameters = 0
    for parameter_group in parameter_groups:
        if isinstance(parameter_group, dict):
            num_parameters += sum(parameter.numel() for parameter in parameter_group['params'])
        else:
            num_parameters += parameter_group.numel()  # type: ignore
    logger.info('Number of trainable parameters: %s', num_parameters)

    # Move the default group to the first, since `modelscope` only log lr of the first group.
    # This is the fastest way I think.
    parameter_groups.reverse()

    return parameter_groups



# Copyright (c) Alibaba, Inc. and its affiliates.


def get_member_set(_class):
    """Get member names set."""
    return set(
        getattr(_class, _a)
        for _a in dir(_class)
        if not _a.startswith('_') and isinstance(getattr(_class, _a), str)
    )


class Tasks:
    """Names for different tasks"""

    word_segmentation = 'word-segmentation'
    part_of_speech = 'part-of-speech'
    named_entity_recognition = 'named-entity-recognition'
    relation_extraction = 'relation-extraction'
    entity_typing = 'entity-typing'
    pretraining = 'pretraining'


class Metrics:
    """Names for different metrics"""

    ner_metric = 'ner-metric'  # alias of sequence-labeling-metric
    sequence_labeling_metric = 'sequence-labeling-metric'
    span_extraction_metric = 'span-extraction-metric'
    relation_extraction_metric = 'relation-extraction-metric'
    typing_metric = 'typing-metric'
    pretraining_metric = 'pretraining-metric'


class Models:
    """Names for different models"""

    sequence_labeling_model = 'sequence-labeling-model'
    global_pointer_model = 'global-pointer-model'
    biaffine_ner_model = 'biaffine-ner-model'
    relation_extraction_model = 'relation-extraction-model'
    multilabel_concat_typing_model = 'multilabel-concat-typing-model'
    multilabel_span_typing_model = 'multilabel-span-typing-model'
    pretraining_model = 'pretraining-model'
    twostage_ner_model = 'twostage-ner-model'


class Embedders:
    """Names for different embedders"""

    embedding = 'embedding'
    transformer_embedder = 'transformer-embedder'


class Encoders:
    """Names for different encoders"""

    span_encoder = 'span-encoder'
    gru_encoder = 'gru'
    rnn_encoder = 'rnn'
    lstm_encoder = 'lstm'


class Decoders:
    """Names for different decoders"""

    crf = 'crf'
    partial_crf = 'partial-crf'
    pairwise_crf = 'pairwise-crf'
    linear = 'linear'
    constrained_crf = 'constrained-crf'


class Preprocessors:
    """Names for different preprocessors"""

    nlp_preprocessor = 'nlp-preprocessor'
    sequence_labeling_preprocessor = 'sequence-labeling-preprocessor'
    relation_extraction_preprocessor = 'relation-extraction-preprocessor'
    span_extraction_preprocessor = 'span-extraction-preprocessor'
    multilabel_span_typing_preprocessor = 'multilabel-span-typing-preprocessor'
    multilabel_concat_typing_preprocessor = 'multilabel-concat-typing-preprocessor'
    pretraining_preprocessor = 'pretraining-preprocessor'
    twostage_preprocessor = 'twostage-preprocessor'


class DataCollators:
    """Names for different data_collators"""

    data_collator_with_padding = 'DataCollatorWithPadding'
    sequence_labeling_data_collator = 'SequenceLabelingDataCollatorWithPadding'
    span_extraction_data_collator = 'SpanExtractionDataCollatorWithPadding'
    multi_label_span_typing_data_collator = 'MultiLabelSpanTypingDataCollatorWithPadding'
    multi_label_concat_typing_data_collator = 'MultiLabelConcatTypingDataCollatorWithPadding'
    pretraining_data_collator = 'PretrainingDataCollatorWithPadding'
    twostage_data_collator = 'TwostageDataCollatorWithPadding'


class Trainers:
    """Names for different trainers"""

    default_trainer = 'default-trainer'


class Optimizers:
    """Names for different optimizers"""

    pass


class LR_Schedulers:
    """Names for different lr_schedulers"""

    pass


class Hooks:
    """Names for different hooks"""

    pass


class DatasetDumpers:
    """Names for different dataset dumpers"""

    ner_dumper = 'ner-dumper'


class Pipelines:
    """Names for different pipelines"""

    span_based_ner_pipeline = 'span-based-ner-pipeline'
    maoe_ner_pipeline = 'maoe-ner-pipeline'



# Copyright (c) Alibaba, Inc. and its affiliates.

from collections import defaultdict
from typing import Dict, List, Set

from modelscope.metrics.base import Metric
from modelscope.metrics.builder import METRICS

from adaseq.data.span_utils import TypedSpan
from adaseq.metainfo import Metrics

UNLABELED_KEY = 'UNLABELED'
MACRO_KEY = 'MACRO'


@METRICS.register_module(module_name=Metrics.span_extraction_metric)
class SpanExtractionMetric(Metric):
    """The metric computation class for span-extraction tasks."""

    def __init__(self, *args, **kwargs):
        self._true_positives: Dict[str, int] = defaultdict(int)
        self._false_positives: Dict[str, int] = defaultdict(int)
        self._false_negatives: Dict[str, int] = defaultdict(int)

    def _update(self, predicted_spans: List[TypedSpan], gold_spans: List[TypedSpan]):
        """
        Update counters.
        """
        gold = set(gold_spans)
        for span in predicted_spans:
            if span in gold:
                self._true_positives[span[2]] += 1
                gold.remove(span)
            else:
                self._false_positives[span[2]] += 1
        # These spans weren't predicted.
        for span in gold:
            self._false_negatives[span[2]] += 1

    def add(self, outputs: Dict, inputs: Dict):
        """Update counters."""
        for predicted_spans, meta in zip(outputs['predicts'], inputs['meta']):
            gold_spans = [(s['start'], s['end'], s['type']) for s in meta['spans']]
            self._update(predicted_spans, gold_spans)
            predicted_unlabeled = [(s[0], s[1], UNLABELED_KEY) for s in predicted_spans]
            gold_unlabeled = [(s[0], s[1], UNLABELED_KEY) for s in gold_spans]
            self._update(predicted_unlabeled, gold_unlabeled)

    def evaluate(self) -> Dict[str, float]:
        """
        Returns:
        `Dict[str, float]`
            A Dict per label containing following the span based metrics:
            - precision : `float`
            - recall : `float`
            - f1 : `float`
            Additionally, an `UNLABELED` key for unlabeled precision, recall and f1,
            a `MACRO` key for macro averaged scores.
        """
        # gather all types, since come counter may have missing keys.
        all_tags: Set[str] = set()
        all_tags.update(self._true_positives.keys())
        all_tags.update(self._false_positives.keys())
        all_tags.update(self._false_negatives.keys())
        # compute per class metrics, including unlabeled scores.
        all_metrics = dict()
        for tag in all_tags:
            all_metrics[tag] = self._compute_metrics(
                self._true_positives[tag], self._false_positives[tag], self._false_negatives[tag]
            )

        # pop unlabeled scores to aggregate macro scores
        unlabeled_metrics = all_metrics.pop(UNLABELED_KEY)
        # macro scores
        macro_p = sum(v['precision'] for v in all_metrics.values()) / len(all_metrics)
        macro_r = sum(v['recall'] for v in all_metrics.values()) / len(all_metrics)
        macro_f1 = 2.0 * (macro_p * macro_r) / (macro_p + macro_r + 1e-13)
        all_metrics[MACRO_KEY] = dict(precision=macro_p, recall=macro_r, f1=macro_f1)
        # put unlabeled scores back
        all_metrics[UNLABELED_KEY] = unlabeled_metrics

        for v in (self._true_positives, self._false_positives, self._false_negatives):
            v.pop(UNLABELED_KEY)
        # overall micro scores
        micro_metrics = self._compute_metrics(
            sum(self._true_positives.values()),
            sum(self._false_positives.values()),
            sum(self._false_negatives.values()),
        )

        scores = micro_metrics
        all_tags.remove(UNLABELED_KEY)
        all_types = [MACRO_KEY, UNLABELED_KEY] + sorted(all_tags)
        for name in all_types:
            for k, v in all_metrics[name].items():
                scores[f'{k}-{name}'] = v
        return scores

    @staticmethod
    def _compute_metrics(
        true_positives: int, false_positives: int, false_negatives: int
    ) -> Dict[str, float]:
        precision = true_positives / (true_positives + false_positives + 1e-13)
        recall = true_positives / (true_positives + false_negatives + 1e-13)
        f1 = 2.0 * (precision * recall) / (precision + recall + 1e-13)
        return dict(precision=precision, recall=recall, f1=f1)



# Copyright (c) Alibaba, Inc. and its affiliates.

from typing import Any, Dict

from modelscope.outputs import OutputKeys
from modelscope.pipelines.builder import PIPELINES
from modelscope.utils.tensor_utils import torch_nested_detach, torch_nested_numpify

from adaseq.metainfo import Pipelines, Tasks

from .base import Pipeline


@PIPELINES.register_module(
    Tasks.named_entity_recognition, module_name=Pipelines.span_based_ner_pipeline
)
class SpanBasedNERPipeline(Pipeline):
    """
    A NER pipeline for span based models, e.g. biaffine model and global pointer model.
    Model outputs should contain:
        predicts: [[(start, end, type), ...]]
    Pipeline outputs will be like:
        {"output": [{"span": "国正", "type": "PER", "start": 0, "end": 2}, ...]}
    """

    def postprocess(  # noqa: D102
        self, inputs: Dict[str, Any], **postprocess_params
    ) -> Dict[str, Any]:
        text = inputs['meta']['text']

        # TODO post_process does not support batch for now.
        offset_mapping = inputs['tokens']['offset_mapping']
        if offset_mapping.dim() == 3:
            offset_mapping = offset_mapping[0]
        offset_mapping = torch_nested_numpify(torch_nested_detach(offset_mapping))

        chunks = inputs['predicts']  # chunks: [[(start, end, type), ...]]
        if isinstance(chunks[0], list):
            chunks = chunks[0]
        chunks = sorted(chunks, key=lambda x: x[0])

        outputs = []
        for chunk in chunks:
            start = offset_mapping[chunk[0]][0]
            end = offset_mapping[chunk[1] - 1][1]
            span_type = chunk[2]
            outputs.append({'span': text[start:end], 'type': span_type, 'start': start, 'end': end})

        return {OutputKeys.OUTPUT: outputs}



# Copyright (c) Alibaba, Inc. and its affiliates.

from typing import Dict, List, Optional, Set

from datasets.arrow_dataset import Dataset

from .span_utils import bio_tags_to_spans


def count_labels(example: Dict, labels: Set, key: str = 'label', **kwargs):
    """
    Count labels from dataset.
    """
    if isinstance(example[key], str):
        labels.add(example[key])
    elif isinstance(example[key], list):
        labels.update(example[key])
    else:
        raise RuntimeError
    return example


def count_span_labels(example: Dict, labels: Set, key: str = 'spans', **kwargs):
    """
    Count labels from dataset.
    """
    for span in example[key]:
        label = span['type']
        if isinstance(label, str):
            labels.add(label)
        elif isinstance(label, list):
            labels.update(label)
        else:
            raise RuntimeError
    return example


def hf_ner_to_adaseq(
    dataset: Dataset,
    key: str = 'ner_tags',
    scheme: str = 'bio',
    classes_to_ignore: Optional[List[str]] = None,
) -> Dataset:
    """
    Map the feature with name `key` from index to label.
    """
    if scheme.lower() == 'bio':
        to_spans = bio_tags_to_spans
    else:
        raise NotImplementedError

    to_str = dataset.features[key].feature.int2str

    def to_adaseq(example):
        tags = [to_str(i) for i in example[key]]
        example['spans'] = to_spans(tags, classes_to_ignore)
        return example

    dataset = dataset.map(to_adaseq)
    return dataset


COUNT_LABEL_FUNCTIONS = {'count_labels': count_labels, 'count_span_labels': count_span_labels}
DATASET_TRANSFORMS = {'hf_ner_to_adaseq': hf_ner_to_adaseq}



# Copyright (c) Alibaba, Inc. and its affiliates.
import logging
import os.path as osp
from functools import partial
from typing import Any, Dict, List, Optional, Union

from datasets import Dataset, DownloadManager
from datasets import load_dataset as hf_load_dataset
from datasets.utils.file_utils import is_remote_url
from modelscope.msdatasets import MsDataset

from adaseq.metainfo import Tasks, get_member_set

from .utils import COUNT_LABEL_FUNCTIONS, DATASET_TRANSFORMS

BUILTIN_TASKS = get_member_set(Tasks)
logger = logging.getLogger(__name__)


class DatasetManager:
    """
    An `DatasetManager` is used by trainers to load a dataset from `modelscope`,
    local files by built-in `DatasetBuilder`s, or custom huggingface-style
    python loading scripts. It finally store a dict object with
    `train`, `valid`, `test` splits of `datasets.Dataset`.

    Args:

    datasets: `Dict[str, Dataset]`, required
        A dict that values are `datasets.Dataset` objects, keys are in `train`,
        `valid`, `test`.
    labels: `Union[str, List[str], Dict[str, Any]]`
        It could be a list of labels, e.g., `['O', 'B-LOC', 'I-LOC', ...]` .
        It could be a path or url to load the label list.
        It could be a kwargs dict to count labels from `train` and `valid` sets,
        e.g., `{'type': 'count_labels', 'key': 'label'}`.
        Types are in `COUNT_LABEL_FUNCTIONS`, you can write your own functions
        and append to `COUNT_LABEL_FUNCTIONS`.
    """

    def __init__(
        self,
        datasets: Dict[str, Dataset],
        labels: Optional[Union[str, List[str], Dict[str, Any]]] = None,
    ) -> None:
        self.datasets = datasets

        if labels is None:
            labels = None
        elif isinstance(labels, list):
            pass
        elif isinstance(labels, str):
            if is_remote_url(labels):
                labels = DownloadManager().download(labels)
            labels = [line.strip() for line in open(labels)]  # type: ignore
        elif isinstance(labels, dict):
            labels = labels.copy()
            label_set = set()
            counter_type = labels.pop('type')
            func = COUNT_LABEL_FUNCTIONS[counter_type]
            counter = partial(func, labels=label_set, **labels)
            kwargs = dict(desc=f'Counting labels by {counter_type}', load_from_cache_file=False)
            datasets['train'].map(counter, **kwargs)
            if 'valid' in datasets:
                datasets['valid'].map(counter, **kwargs)
            labels = sorted(label_set)
        else:
            raise ValueError(f'Unsupported labels: {labels}')
        self.labels = labels

    @property
    def train(self):  # noqa: D102
        return self.datasets.get('train', None)

    @property
    def dev(self):  # noqa: D102
        return self.datasets.get('valid', None)

    @property
    def valid(self):  # noqa: D102
        return self.datasets.get('valid', None)

    @property
    def test(self):  # noqa: D102
        return self.datasets.get('test', None)

    @classmethod
    def from_config(
        cls,
        task: Optional[str] = None,
        data_file: Optional[Union[str, Dict[str, str]]] = None,
        path: Optional[str] = None,
        name: Optional[str] = None,
        access_token: Optional[str] = None,
        labels: Optional[Union[str, List[str], Dict[str, Any]]] = None,
        transform: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> 'DatasetManager':
        """
        load dataset and construct an `DatasetManager`.

        Args:

        task: `str`, optional (default = `None`)
            Specifies the task when loading a dataset with built-in `DatasetBuilder`s.
            Now we have `NamedEntityRecognitionDatasetBuilder` and
            `EntityTypingDatasetBuilder`.
            It could also be the absolute path of a custom python script for the
            `datasets.load_dataset`.

        data_file: `Union[str, Dict[str, str]]`, optional (default = `None`)
            Used only when loading a dataset with built-in `DatasetBuilder`s.
            It could be an url like `"https://data.deepai.org/conll2003.zip"`,
            or a local directory (absolute path) like `"/home/data/conll2003"`,
            or a local archive file absolute path like `"/home/data/conll2003.zip"`.
            Details refer to `./dataset_builders/base.py`.
            It could also be a dict that specifies file paths (absolute) for each
            dataset splits, for example:
            `{'train': '/home/data/train.txt', 'valid': '/home/data/valid.txt'}`, or
            `{'train': 'http://DOMAIN/train.txt', 'valid': 'http://DOMAIN/valid.txt'}`.

        path: `str`,  optional (default = `None`)
            It is the name of a huggingface-hosted dataset or the absolute path
            of a custom python script for the `datasets.load_dataset`.

        name: `str`,  optional (default = `None`)
            If `data_file` or `path` are given, it is the `name` argument in
            `datasets.load_dataset`.
            If not, it is used as a modelscope dataset name.

        access_token: `str`, optional (default = `None`)
            If given, use this token to login modelscope, then we can access some
            private datasets.

        transform: `Dict[str, Any]` (default = `None`)
            function kwargs for reformat datasets, see `apply_transform()`.
        """
        if data_file is not None:
            if isinstance(data_file, str):
                if not is_remote_url(data_file) and not osp.exists(data_file):
                    raise RuntimeError('`data_file` not exists: %s', data_file)
                kwargs.update(data_dir=data_file)
            elif isinstance(data_file, dict):
                for k, v in data_file.items():
                    if is_remote_url(v):
                        continue
                    if not osp.exists(v):
                        raise RuntimeError('`data_file[%s]` not exists: %s', k, v)
                    if not osp.isabs(v):
                        # since datasets
                        raise RuntimeError('`data_file[%s]` must be a absolute path: %s', k, v)

                # we rename all `dev` key to `valid`
                if 'dev' in data_file:
                    data_file['valid'] = data_file.pop('dev')
                kwargs.update(data_files=data_file)
            else:
                raise ValueError(f'Unsupported data_file: {data_file}')

            assert task is not None and task in BUILTIN_TASKS, 'Need a specific task!'
            # where we have some pre-defined dataset builders
            task_builder_name = task
            if task in {Tasks.word_segmentation, Tasks.part_of_speech}:
                task_builder_name = Tasks.named_entity_recognition

            path = osp.join(
                osp.dirname(osp.abspath(__file__)),
                'dataset_builders',
                task_builder_name.replace('-', '_') + '_dataset_builder.py',
            )

        if isinstance(path, str):
            if path.endswith('.py') or osp.isdir(path):
                logger.info('Will use a custom loading script: %s', path)

            if name is not None:
                logger.info("Passing `name='%s'` to `datasets.load_dataset`", name)

            hfdataset = hf_load_dataset(path, name=name, **kwargs)
            datasets = {k: v for k, v in hfdataset.items()}

        elif isinstance(name, str):
            # to access private datasets from modelscope
            if access_token is not None:
                from modelscope.hub.api import HubApi

                HubApi().login(access_token)

            msdataset = MsDataset.load(name, **kwargs)
            datasets = {k: v._hf_ds for k, v in msdataset.items()}

        else:
            raise RuntimeError('Unsupported dataset!')

        if 'dev' in datasets and 'valid' not in datasets:
            datasets['valid'] = datasets.pop('dev')

        if 'test' in datasets and 'valid' not in datasets:
            datasets['valid'] = datasets['test']

        if 'valid' in datasets and 'test' not in datasets:
            datasets['test'] = datasets['valid']

        # apply transform
        if transform:
            datasets = apply_transform(datasets, **transform)

        if labels is None:
            if task in {
                Tasks.named_entity_recognition,
                Tasks.word_segmentation,
                Tasks.part_of_speech,
                Tasks.entity_typing,
            }:
                labels = dict(type='count_span_labels')
            elif task == Tasks.relation_extraction:
                labels = dict(type='count_labels')

        if 'train' in datasets:
            logger.info('First sample in train set: ' + str(datasets['train'][0]))

        return cls(datasets, labels)  # type: ignore


def apply_transform(
    datasets: Dict[str, Dataset],
    name: str,
    **kwargs,
) -> Dict[str, Dataset]:
    """
    Apply a given function to reformat a `Dataset`.
    """
    if name not in DATASET_TRANSFORMS:
        raise RuntimeError(f'{name} not in {DATASET_TRANSFORMS.keys()}')
    kwargs = kwargs or dict()

    for k in datasets.keys():
        datasets[k] = DATASET_TRANSFORMS[name](datasets[k], **kwargs)

    return datasets



# Copyright (c) Alibaba, Inc. and its affiliates.

import json

import datasets
from datasets import Features, Value

from .base import CustomDatasetBuilder


class EntityTypingDatasetBuilderConfig(datasets.BuilderConfig):
    """BuilderConfig for entity typing datasets"""

    def __init__(self, data_dir=None, data_files=None, **corpus_config):
        super().__init__(data_dir=data_dir, data_files=data_files)
        self.corpus_config = corpus_config


class EntityTypingDatasetBuilder(CustomDatasetBuilder):
    """Builder for entity typing datasets.

    features:
        id: string, data record id.
        tokens: list[str] input tokens.
        spans: List[Dict],  mentions like: [{'start': 0, 'end': 2, 'type': ['PER', 'MAN']}]
        mask: bool, mention mask.
    """

    BUILDER_CONFIG_CLASS = EntityTypingDatasetBuilderConfig

    def stub():  # noqa: D102
        pass

    def _info(self):
        info = datasets.DatasetInfo(
            features=Features(
                {
                    'id': Value('string'),
                    'tokens': [Value('string')],
                    'spans': [
                        {
                            'start': Value('int32'),  # close
                            'end': Value('int32'),  # open
                            'type': [Value('string')],
                        }
                    ],
                    'mask': [Value('bool')],
                }
            )
        )
        return info

    def _generate_examples(self, filepath):
        """Load data file.
        Args:
        file_path: string, data file path.
        corpus_config: dictionary, required keys:
        tokenizer: string, specify the tokenization method,
            'char' list(text) and 'blank' for text.split(' ')
        is_end_included: bool, wheather the end index pointer to the
            last token or the token next to the last token.
            e.g., text = 'This is an example.', mention = 'example',
            end = 3 if is_end_included is True.

        """
        corpus_config = self.config.corpus_config
        spans_key = corpus_config.get('spans_key', 'label')
        text_key = corpus_config.get('text_key', 'text')
        tokenizer = corpus_config.get('tokenizer', 'char')
        is_end_included = corpus_config.get('is_end_included', False)

        with open(filepath, encoding='utf-8') as f:
            guid = 0
            for line in f:
                example = json.loads(line)
                text = example[text_key]
                if isinstance(text, list):
                    tokens = text
                elif isinstance(text, str):
                    if tokenizer == 'char':
                        tokens = list(text)
                    elif tokenizer == 'blank':
                        tokens = text.split()
                    else:
                        raise NotImplementedError
                else:
                    raise ValueError('Unsupported text input.')

                entities = list()
                for span in example[spans_key]:
                    if is_end_included:
                        span['end'] += 1
                    entities.append(span)

                mask = [True] * len(tokens)

                reading_format = corpus_config.get('encoding_format', 'span')
                if reading_format == 'span':
                    yield guid, {
                        'id': str(guid),
                        'tokens': tokens,
                        'spans': entities,
                        'mask': mask,
                    }
                    guid += 1
                elif reading_format == 'concat':
                    for entity in entities:
                        yield guid, {
                            'id': str(guid),
                            'tokens': tokens,
                            'spans': [entity],
                            'mask': mask,
                        }
                        guid += 1
                else:
                    raise NotImplementedError('unimplemented reading_format')
